
var P = require('bluebird');
var _ = require('lodash');
var injector = require('./injector');
var Joi = require('joi');
var fs = require('fs');
var random = require('meteor-random');

module.exports = function(config) {
    var log = require('./log')(config);

    log.info("Instantiating the platform");

    var FN_ARGS = /^function\s*[^\(]*\(\s*([^\)]*)\)/m;
    var FN_ARG_SPLIT = /,/;
    var STRIP_COMMENTS = /((\/\/.*$)|(\/\*[\s\S]*?\*\/))/mg;

    var seneca = require('seneca')({
        log: 'silent'
    });

    function getParamNames(fn) {
        var fnText = fn.toString().replace(STRIP_COMMENTS, '');
        var argDecl = fnText.match(FN_ARGS);
        return argDecl[1].split(FN_ARG_SPLIT);
    }


    var _validateMessage = P.method(function(service, msg) {
        if(service.schema) {
            return P.promisify(Joi.validate, {context:Joi})(msg, service.schema, { allowUnknown: service.allowUnknown || true, stripUnknown: service.stripUnknown || true });
        }
        else {
            return msg;
        }
    });

    seneca.actAsync = P.promisify(seneca.act, {context:seneca});

    function _registerServices(serviceRoot) {
        log.debug("Register services", serviceRoot);
        return P.method(function(platform) {
            if(serviceRoot) {
                serviceRoot = require("path").resolve(serviceRoot);
                log.debug("Service root", serviceRoot);

                return new P(function(resolve) {
                    try {
                        fs.statSync(serviceRoot).isDirectory();
                        log.debug("Loading all services from folder %s", serviceRoot);
                        resolve(P.map(fs.readdirSync(serviceRoot), function(file) {
                            log.debug("Registering service",file);
                            if(file.indexOf('.service') !== -1) {
                                log.debug("Loading service", file);
                                var factory = require(serviceRoot + "/" + file);
                                var params = getParamNames(factory);
                                var args = [];
                                _.forEach(params, function(param) {
                                    args.push(injector.resolve(param));
                                });
                                var spec = factory.apply(platform, args);

                                // Register this service with seneca
                                seneca.add(spec.pattern, function(msg, done) {
                                    return _validateMessage(spec, msg).then(function(msg) {
                                        return P.method(spec.handler)(msg).then(function(resp) {
                                            // Apply any responseSchema before
                                            if(spec.responseSchema) {
                                                return P.promisify(Joi.validate, {context: Joi})(resp, spec.responseSchema, { allowUnknown: spec.allowUnknown || true, stripUnknown: spec.stripUnknown || true }).catch(function(err){
                                                    log.warn("Validation Error:", err);
                                                    log.warn("Context is", service);
                                                    throw err;
                                                });
                                            }
                                            else {
                                                return resp;
                                            }
                                        }).asCallback(done);
                                    }).catch(function(err) {
                                        log.error("ERROR [%s]:"+spec.pattern, err);
                                        done(err);
                                    });
                                });

                                // Keep all specs for potential future use
                                return {name: file, spec: spec};
                            }
                        }).then(function(services) {
                            log.debug("Platform instance was successfully started with %d service(s)", services.length);
                            platform.services = services;
                            return platform;
                        }));
                    }
                    catch(e) {
                        log.warn("No services to load");
                        resolve(platform);
                    }

                });
            }
            else {
                return platform;
            }
        })
    }

    return {
        startService: function (serviceRoot = "./src/services", options = {}) {
            log.info("Starting service");
            return new P((resolve, reject) => {

                // Register globally accessible helpers
                injector.register('config', config);
                injector.register('log', log);
                injector.register('injector', injector);
                injector.register('platform', this);
                injector.register('seneca', seneca);
                injector.register('Promise', P);
                injector.register('_', _);
                injector.register('Schema', Joi);
                injector.register('Random', random);
                injector.register('Credentials', require('../schemas/credentials.schema')(this));
                injector.register('Clock', require('./clock'));
                injector.register('Calendar', require('./calendar'));
                injector.register('Boom', require('boom'));
                if(config.db !== false) {
                    injector.register('DBS', require('./db')(config, log, options.testMode));
                }
                injector.register('ModelFactory', require('./model-factory')(this, options));

                // Inject any additional elements
                if(options.inject && options.inject.early) {
                    _.forEach(options.inject.early, (value, key) => {
                        if(_.isFunction(value)) {
                            injector.register(key, value(this));
                        }
                        else {
                            injector.register(key, value);
                        }
                    });
                }

                if (!options.testMode) {

                    if(config.gateway) {
                        // Make sure we connect to the gateway
                        seneca.client({type: 'tcp', port: config.gateway.port, host: config.gateway.host, pin: 'srv: gateway, tar: *, act: *'});
                    }

                    if(config.http) {
                        log.debug("Connecting HTTP %s", config.http.host + ":" + config.http.port);
                        seneca.listen(config.http, function(err) {
                            if(err) {
                                log.error("Unable to establish HTTP protocol", err);
                            }
                            else {
                                log.info("Successfully connected HTTP protocol");
                            }
                        });
                    }

                    if(config.tcp) {
                        seneca.listen({type: 'tcp', port: config.tcp.port, pin: config.tcp.pin }, function(err) {
                            if(err) {
                                log.error("Unable to establish TCP connection on port ", config.tcp.port);
                            }
                            else {
                                log.info("Listening for incoming messages on TCP %d", config.tcp.port);
                            }
                        });
                    }

                    if(config.amqp) {
                        log.debug("Connecting to AMQP %s", config.amqp.url);
                        seneca.use('seneca-amqp-transport', {amqp: config.amqp}).listen({ type: 'amqp', name: config.amqp.name, pin: config.amqp.pin }, (err) => {
                            if (err) {
                                log.error(err);
                                reject(err);
                            }
                            else {
                                log.debug("Successfully connected to AMQP server");
                            }
                        });

                        seneca.client({type: 'amqp'});
                    }

                    if(config.redis) {
                        log.debug("Connecting to Redis transport", config.redis);
                        seneca.use('redis-queue-transport', { 'redis-queue': {
                            timeout: config.redis.timeout || 30000,
                            type: 'redis-queue',
                            host: config.redis.host,
                            port: config.redis.port
                        }}).client({type: 'redis-queue'}).listen({type: 'redis-queue', pin: config.redis.pin}, (err) => {
                            if(err) {
                                log.error(err);
                            }
                            else {
                                log.info("Successfully listening through redis channel using pin: ", config.redis.pin);
                            }

                        });
                    }

                    resolve(this);
                }
                else {
                    log.warn("Not connecting any external transports while testing");
                    resolve(this);
                }

            }).then(function(platform) {

                // Inject any additional elements
                if(options.inject && options.inject.late) {
                    _.forEach(options.inject.late, (value, key) => {
                        if(_.isFunction(value)) {
                            injector.register(key, value(platform));
                        }
                        else {
                            injector.register(key, value);
                        }
                    });
                }

                return platform;
            }).then(_registerServices(serviceRoot));
        },
        getInjector: function () {
            return injector;
        },
        execute: function (pattern, params, options = {}) {
            return seneca.actAsync.apply(seneca.actAsync, arguments);
        }
    };
};
